# Copyright 2016 Greg M. Bernstein
"""
This is a basic application to set up source-destination explicitly routed
paths based on Ethernet addresses.

This application will only work with Mininet networks whose topologies are known
in advanced and available in a JSON file. Substitute your network file for my
ExNetweithLoops1A.json file.

To launch the application just use:
    python EthPathsBasic.py --netfile=ExNetwithLoops1A.json
If you don't want to use the telnet/python backdoor:
    python EthPathsBasic.py --netfile=ExNetwithLoops1A.json --notelnet

To launch mininet using a custom topology and remote controller based on this
code use:

sudo python NetRunnerNS.py -f ExNetwithLoops1A.json -ip address_of_this_code

To bring up our 'backdoor' python interface on the machine you are running this
program (windows or Linux) type:
telnet localhost 3000

Getting a reference to the Outline application in the 'backdoor' python shell
and setting up and tearing down a path:
```python
from ryu.base.app_manager import AppManager
am = AppManager.get_instance()
myapp = am.applications["EthPathsBasic"]
myapp.make_path("H2-S1-S2-H5") # sets up a path
myapp.make_path("H2-S1-S2-H5", delete=True) # Deletes the path
```
"""
from builtins import range
import json
from networkx.readwrite import json_graph
from ryu.base import app_manager
from ryu.controller import ofp_event
from ryu.controller.handler import set_ev_cls
from ryu.cmd import manager  # For directly starting Ryu
import sys  # For getting command line arguments and passing to Ryu
import eventlet
from eventlet import backdoor  # For telnet python access
from ryu.ofproto import ofproto_v1_0  # This code is OpenFlow 1.0 specific

if __name__ == "__main__":  # Stuff to set additional command line options
    from ryu import cfg
    CONF = cfg.CONF
    CONF.register_cli_opts([
        cfg.StrOpt('netfile', default=None, help='network json file'),
        cfg.BoolOpt('notelnet', default=False,
                    help='Telnet based debugger.')
    ])


class EthPathsBasic(app_manager.RyuApp):
    """ Outline of a Ryu application that performs a number of tasks and
        responds to various events.
    """
    OFP_VERSIONS = [ofproto_v1_0.OFP_VERSION]

    def __init__(self, *args, **kwargs):
        """ Define member variables here.  I do this here so that we can look at
            them via the telnet python command line.  I use the standard Python
            deque data structure with a finite length so that I only keep the
            latest few events of a given type.
        """
        super(EthPathsBasic, self).__init__(*args, **kwargs)
        self.netfile = self.CONF.netfile
        self.switches = {}
        self.g = json_graph.node_link_graph(json.load(open(self.netfile)))
        if not self.CONF.notelnet:
            eventlet.spawn(backdoor.backdoor_server,
                           eventlet.listen(('localhost', 3000)))

    @set_ev_cls(ofp_event.EventOFPSwitchFeatures)
    def switch_features(self, event):
        """ This method gets called when a switch connects to the controller.
        """
        msg = event.msg
        dp = msg.datapath
        # Assumes that datapath ID represents an ascii name
        switchName = dpidDecode(dp.id)
        self.logger.info("Switch {} came up".format(switchName))
        # self.logger.info("Message: {}".format(msg))
        self.switches[switchName] = dp  # Save switch information


    def make_path(self, pathString, delete=False):
        """ Sets up a bi-directional path between hosts based on the pathString.
        The pathString consists of node names separated by - (dashes) with the
        first and last node being switchs and all other nodes being switches.
        """
        self.logger.info("make_path called with path {}".format(pathString))
        node_list = pathString.split("-")
        if not path_valid(self.g, node_list):
            self.logger.info("Invalid path, cannot create!")
            return
        src = node_list[0]
        dst = node_list[-1]
        src_MAC = self.g.node[src]['mac']
        dst_MAC = self.g.node[dst]['mac']
        # Setup the forward path
        for i in range(1, len(node_list) - 1):
            switch = node_list[i]
            out_port = get_out_port(self.g, node_list[i], node_list[i + 1])
            datapath = self.switches[switch]
            ofproto = datapath.ofproto
            parser = datapath.ofproto_parser
            if not delete:
                command = ofproto.OFPFC_ADD
            else:
                command = ofproto.OFPFC_DELETE
            actions = [parser.OFPActionOutput(out_port)]
            match = parser.OFPMatch(dl_dst=dst_MAC, dl_src=src_MAC)
            mod = parser.OFPFlowMod(
                datapath=datapath, match=match, cookie=0,
                command=command, idle_timeout=0, hard_timeout=0,
                priority=20,
                flags=ofproto.OFPFF_SEND_FLOW_REM, actions=actions)
            datapath.send_msg(mod)

        # Setup the reverse path
        node_list.reverse()
        for i in range(1, len(node_list) - 1):
            switch = node_list[i]
            out_port = get_out_port(self.g, node_list[i], node_list[i + 1])
            datapath = self.switches[switch]
            ofproto = datapath.ofproto
            parser = datapath.ofproto_parser
            if not delete:
                command = ofproto.OFPFC_ADD
            else:
                command = ofproto.OFPFC_DELETE
            actions = [parser.OFPActionOutput(out_port)]
            # Switch addresses for reverse path
            match = parser.OFPMatch(dl_dst=src_MAC, dl_src=dst_MAC)
            mod = parser.OFPFlowMod(
                datapath=datapath, match=match, cookie=0,
                command=command, idle_timeout=0, hard_timeout=0,
                priority=20,
                flags=ofproto.OFPFF_SEND_FLOW_REM, actions=actions)
            datapath.send_msg(mod)


def path_valid(g, p):
    """ Checks whether the list nodes p is a valid path in g."""
    plen = len(p)
    for i in range(plen - 1):
        if not g.has_edge(p[i], p[i + 1]):
            return False
    return True

def get_out_port(g, n1, n2):
    """ Get the output port for n1 to send to n2
        assuming n1 and n2 are connected and ports are assigned to links."""
    # Uses nifty NetworkX graph representation.
    return g[n1][n2]["ports"][n1]


def dpidDecode(aLong):
    try:
        myBytes = bytearray.fromhex('{:8x}'.format(aLong)).strip()
        return myBytes.decode()
    except ValueError:
        return str(aLong)

            
if __name__ == "__main__":
    manager.main(args=sys.argv)
